##
## Simulation from Pashley & Miratrix
## Insights on Variance Estimation for Blocked and Matched Pairs Designs
##
## Using GitHub blkvar package
## https://github.com/lmiratrix/blkvar/releases/tag/insights
##
## (C) 2020 N. Pashley
##

library( blkvar )
library( ggplot2 )
library( tidyverse )
library( ggthemes )

# Core simulation function
# data - schedule of potential outcomes
# p.mat - list of proportions of treated units in each block
# nsims - number of replicates.
simFun<-function( data, p.mat, nsims ){
  data$B = as.numeric( data$B )

  # Use the full schedule of potential outcomes to calculate exact variances
  # using theory.
  # Used as a sanity check of the math formula.

  #This version of the function give correct number treated+control but uses all POs to calculate values
  data.table<-blkvar::calc_summary_stats_oracle(Y0, Y1, B, data=data, p.mat)

  #Get true design based variances. This also puts true potential outcomes into
  #regression methods to get "population regression" values
  v1<-compare_methods_oracle(Y0, Y1, B, data=data, p.mat)

  #Calculate true variance under complete randomization
  s.tc.cr<-var((data$Y1-data$Y0))
  cr_var<-var(data$Y1)/sum(data.table$n1) + var(data$Y0)/sum(data.table$n0)-s.tc.cr/sum(data.table$n1+data.table$n0)
  #Calculate true variance under block randomization
  s.tc.bk<-aggregate(list(s.tc.bk=data$Y1-data$Y0), list(B=data$B), FUN=var)
  overall_table<-merge(data.table, s.tc.bk, by="B" )
  overall_table$n_k<-overall_table$n1+overall_table$n0
  bk_var<-sum((overall_table$n_k)^2*((overall_table$se_ney)^2-overall_table$s.tc.bk/(overall_table$n_k))/sum(overall_table$n_k)^2)
  #Calculate a metric of within vs between variance
  r_1<-by(data$Y1, data$B, function(x) blkvar:::within_blk_var(x) )
  r_2<-by(data$Y0, data$B, function(x) blkvar:::within_blk_var(x) )
  r_3<-sum((data$Y1-mean(data$Y1))^2)
  r_4<-sum((data$Y0-mean(data$Y0))^2)
  R_2<- 1-(sum(r_1)+sum(r_2))/(r_3+r_4)

  #now get estimates based on simulations
  v2<-as.list( rep( NA, nsims ) )
  n<-sum(overall_table$n_k)
  tau.cr.est<-vector(length=nsims)
  var.cr.est<-vector(length=nsims)
  for(i in 1:nsims){
    #First randomly assign units to treatments using block randomization
    temp.data<-data.frame(B=data$B)
    temp.data$index<-1:length(data$B)
    temp.data<-merge(temp.data, overall_table, by="B")
    trt.bk<-unlist(sapply(split(temp.data, temp.data$B),
                          function(x) sample(x[,"index"], size=x[,"n1"])),use.names=FALSE)
    Z.bk<-rep(0,n)
    Z.bk[trt.bk]<-1
    Y<-data$Y0*(1-Z.bk)+data$Y1*Z.bk

    #Now get all the variance estimates
    simdat<-data.frame(Yobs = Y,
                     Z = Z.bk,
                     B = data$B )
    v2[[i]] = compare_methods(Yobs=Yobs, Z=Z, B=B, data=simdat,
                                  include_MLM = FALSE, include_DB = FALSE,
                                  weight_LM_method = c("precision"))

    #Compare to complete randomization
    Z.cr<-sample( simdat$Z )
    Y.trt<-data$Y1[Z.cr == 1]
    Y.ctrl<-data$Y0[Z.cr == 0]
    tau.cr.est[i]<-mean(Y.trt)-mean(Y.ctrl)
    var.cr.est[i]<-var(Y.trt)/length(Y.trt)+var(Y.ctrl)/length(Y.ctrl)
  }
  v2 = bind_rows(v2)

  #Collect, name, and return values
  return_val<-list(v1, v2, bk_var, cr_var, tau.cr.est, var.cr.est, R_2)
  names(return_val)<-c("v1", "v2", "bk_var", "cr_var", "tau.cr.est", "var.cr.est", "R_2")
  return(return_val)
}



#### Simulations #####

##Initialize values
set.seed(73264782)
nsims<-5000

## Fixed parameters (block sizes, etc.)
#Number of units in each block
n_k_set<-c(3, 3, 3, 3, 4, 4, 7, 7, 7, 7, 7, 10, 10, 15, 20)
#Proportion treated in each block
p_set<-c(1/3, 1/3, 1/3, 1/3, 1/4, 1/4, 1/7, 1/7, 1/7, 1/7, 1/7, 1/5, 1/5, 1/5, 1/5)
#Total number of blocks
K<-length(n_k_set)

# Matrix of treatment proportions
p_mat<-data.frame( B = 1:K,
                   p=p_set )
p_mat$B<-as.character(p_mat$B)

#Variance of [treated and potential outcomes
sigma_c1<-rep(1,K)
sigma_t1<-rep(1,K)


## Set things that will change on each run
percents<-seq(from=(1-1/(K+1)), to=1/(K+1), by=-1/(K+1))
#Separation of control means across blocks
alpha_1<-qnorm(percents, 0, 0)
alpha_2<-qnorm(percents, 0, 0.1)
alpha_3<-qnorm(percents, 0, 0.3)
alpha_4<-qnorm(percents, 0, 0.5)
alpha_5<-qnorm(percents, 0, 0.8)
alpha_6<-qnorm(percents, 0, 1)
alpha_7<-qnorm(percents, 0, 1.5)
alpha_8<-qnorm(percents, 0, 2)
alpha_all<-cbind(alpha_1, alpha_2, alpha_3,alpha_4,alpha_5, alpha_6, alpha_7, alpha_8)
alpha_sd_all<-c(0,0.1,0.3,0.5,0.8,1,1.5,2)
alpha_sd_exact<-c(sd(alpha_1), sd(alpha_2), sd(alpha_3),
                  sd(alpha_4), sd(alpha_5), sd(alpha_6),
                  sd(alpha_7), sd(alpha_8))
alpha_sd_exact<-round(alpha_sd_exact,3)
alpha_sd_exact

#Sepation of treatment effects across blocks
beta_1<-qnorm(percents, 5, 0)
beta_2<-qnorm(percents, 5, 0.1)
beta_3<-qnorm(percents, 5, 0.3)
beta_4<-qnorm(percents, 5, 0.5)
beta_5<-qnorm(percents, 5, 0.8)
beta_6<-qnorm(percents, 5, 1)
beta_7<-qnorm(percents, 5, 1.5)
beta_8<-qnorm(percents, 5, 2)
beta_all<-cbind(beta_1, beta_2, beta_3, beta_4, beta_5, beta_6, beta_7, beta_8)
beta_sd_all<-c(0,0.1,0.3,0.5,0.8,1,1.5,2)
beta_sd_exact<-c(sd(beta_1), sd(beta_2), sd(beta_3),
                 sd(beta_4), sd(beta_5), sd(beta_6),
                 sd(beta_7), sd(beta_8))
beta_sd_exact<-round(beta_sd_exact,3)

#Correlation of potential outcomes within each block
rho_1<-rep(0,K)
rho_2<-rep(0.5,K)
rho_3<-rep(1,K)
rho_all<-cbind(rho_1, rho_2, rho_3)
rho_val_all<-c(0,0.5,1)

#### Run the Simulations ####


if ( FALSE ) {
  # test simulation function.
  Y.temp<-generate_blocked_data(n_k_set, sigma_alpha = 0.1, sigma_tau = 0.5, tau = 5,
                    sigma_0 = 1, sigma_1 = 1, corr = 0.5, exact=TRUE )
  head( Y.temp )

  output_sims<-simFun(Y.temp, p_mat, 100)
  names( output_sims )
  output_sims$v1
  head( output_sims$v2 )
  nrow( output_sims$v2 )
  output_sims$bk_var
  output_sims$cr_var
  output_sims$tau.cr.est
  output_sims$var.cr.est
  output_sims$R_2
}



v1_out_save<-data.frame()
v2_out_save<-data.frame()
bk_vs_cr_out1<-data.frame()
bk_vs_cr_out2<-data.frame()
s<-0
t<-0
for(j in 1:length(alpha_sd_all)){
  print(j)
  for(i in 1:length(beta_sd_all)){
    t<-t+1
    for(k in 1:length(rho_val_all)){
      s<-s+1
      print(s)

      #Generate data
      Y.temp<-generate_blocked_data(n_k_set,
                        sigma_alpha = alpha_sd_all[j],
                        sigma_beta = beta_sd_all[i],
                        sigma_0 = 1, sigma_1 = 1,
                        corr = rho_val_all[k], exact=TRUE )

      #Run simulation
      output_sims<-simFun(Y.temp, p_mat, nsims)

      #Save outputs
      v1.temp<-output_sims$v1
      temp.val<-nrow(v1.temp)
      v1.temp$alpha<-rep(alpha_sd_exact[j], temp.val)
      v1.temp$beta<-rep(beta_sd_exact[i], temp.val)
      v1.temp$rho<-rep(rho_val_all[k], temp.val)
      v1_out_save<-rbind(v1_out_save, v1.temp)
      v2.temp<-output_sims$v2
      temp.val<-nrow(v2.temp)
      v2.temp$alpha<-rep(alpha_sd_exact[j], temp.val)
      v2.temp$beta<-rep(beta_sd_exact[i], temp.val)
      v2.temp$rho<-rep(rho_val_all[k], temp.val)
      v2_out_save<-rbind(v2_out_save, v2.temp)
      bk_var_temp<-output_sims$bk_var
      cr_var_temp<-output_sims$cr_var
      tau.cr.est_temp<-output_sims$tau.cr.est
      var.cr.est_temp<-output_sims$var.cr.est
      temp.frame1<-data.frame(bk_var_temp)
      names(temp.frame1)<-c("bk_var")
      temp.frame1$cr_var<-cr_var_temp
      temp.frame2<-data.frame(tau.cr.est_temp)
      names(temp.frame2)<-c("tau.cr.est")
      temp.frame2$var.cr.est<-var.cr.est_temp
      temp.val<-nrow(temp.frame1)
      temp.frame1$alpha<-rep(alpha_sd_exact[j], temp.val)
      temp.frame1$beta<-rep(beta_sd_exact[i], temp.val)
      temp.frame1$rho<-rep(rho_val_all[k], temp.val)
      temp.frame1$R_2<-rep(output_sims$R_2, temp.val)
      temp.val<-nrow(temp.frame2)
      temp.frame2$alpha<-rep(alpha_sd_exact[j], temp.val)
      temp.frame2$beta<-rep(beta_sd_exact[i], temp.val)
      temp.frame2$rho<-rep(rho_val_all[k], temp.val)
      bk_vs_cr_out1<-rbind(bk_vs_cr_out1, temp.frame1)
      bk_vs_cr_out2<-rbind(bk_vs_cr_out2, temp.frame2)
    }
  }
}

#Save data
write.csv(v1_out_save, file = "v1_out_save.csv",row.names=FALSE)
write.csv(v2_out_save, file = "v2_out_save.csv",row.names=FALSE)
write.csv(bk_vs_cr_out1, file = "bk_vs_cr_out1.csv",row.names=FALSE)
write.csv(bk_vs_cr_out2, file = "bk_vs_cr_out2.csv",row.names=FALSE)


#### Analyze simulation results #####

#Read data
v1_out_save<-read.csv(file = "v1_out_save.csv", header=TRUE)
v2_out_save<-read.csv(file = "v2_out_save.csv", header=TRUE)
bk_vs_cr_out1<-read.csv(file = "bk_vs_cr_out1.csv", header=TRUE)
bk_vs_cr_out2<-read.csv(file = "bk_vs_cr_out2.csv", header=TRUE)

### Check that our formulas match simulation estimates for variance estimator
### expectation

#Simulation
v2_out_save$var_est<-(v2_out_save$SE)^2
avg.val.mat<-v2_out_save%>%group_by(alpha,beta, rho, method) %>%
  summarise(var_est = mean(var_est) )
#Formula
v1_out_save$var_est_exp<-v1_out_save$SE^2
v1_out_save$method<-v1_out_save$Methods
combine.mat<-merge(avg.val.mat, v1_out_save, by=c("alpha", "beta", "rho", "method"))
combine.mat$Diff<-combine.mat$var_est-combine.mat$var_est_exp
#Differences
summary(combine.mat$Diff[which(combine.mat$method=="hybrid_p")])
summary(combine.mat$Diff[which(combine.mat$method=="hybrid_m")])

#Check completely randomized variance looks correct
cr.var.val.mat<-bk_vs_cr_out2%>%group_by(alpha,beta, rho) %>%
  summarise(Var_exp = var(tau.cr.est) )
combine.mat2<-merge(cr.var.val.mat, bk_vs_cr_out1, by=c("alpha", "beta", "rho"))
combine.mat2$Diff_cr<-combine.mat2$Var_exp-combine.mat2$cr_var
summary(combine.mat2$Diff_cr)
summary(combine.mat2$Var_exp)
summary(combine.mat2$cr_var)

#Check blocking variance looks correct
temp.mat<-v2_out_save[which(v2_out_save$method=="hybrid_m"),]
bk.var.val.mat<-temp.mat%>%group_by(alpha,beta, rho) %>%
  summarise(Var_exp = var(ATE_hat) )
combine.mat3<-merge(bk.var.val.mat, bk_vs_cr_out1, by=c("alpha", "beta", "rho"))
combine.mat3$Diff_bk<-combine.mat3$Var_exp-combine.mat3$bk_var
summary(combine.mat3$Diff_bk)
summary(combine.mat3$Var_exp)
summary(combine.mat3$bk_var)


#### Plot simulation results #####

##First look at Relative Bias
var.est.frame<-merge(avg.val.mat, bk_vs_cr_out1, by=c("alpha", "beta", "rho"))
var.est.frame$Bias<-var.est.frame$var_est-var.est.frame$bk_var
var.est.frame$Relative_Bias1<-var.est.frame$Bias/var.est.frame$bk_var
var.est.frame$Relative_Bias<-var.est.frame$Relative_Bias1*100


var.est.frame1<-filter( var.est.frame,
                        method %in% c( "hybrid_m","hybrid_p","FE-Het","FE-IPTW(pr)") )


head( var.est.frame1 )

table( var.est.frame1$alpha )

theme_set( theme_tufte() )
s.var.est = var.est.frame1 %>%
  group_by( method, beta, rho ) %>%
  dplyr::summarize( n = n(),
                    Relative_Bias = mean(Relative_Bias) )

ggplot(s.var.est,
        aes( x=beta, y=Relative_Bias, group= method ) ) +
  facet_grid( . ~ rho, labeller = label_both)+
  geom_hline( yintercept=0, col="grey" ) +
  geom_line(aes(linetype=method, color=method)) + geom_point(aes(shape=method,color=method), size=1) +
  coord_cartesian( ylim=c(-20,230 ) )  +
  labs( y = "Estimation Inflation (%)", x ="SD of Block Average Treatment Effects" )+
  scale_colour_discrete(name  ="Methods",
                        breaks=c("FE-Het", "hybrid_m", "hybrid_p", "FE-IPTW(pr)"),
                        labels=c("Fixed effects", 
                                 expression('Hybrid'['m']), expression('Hybrid'['p']), "Weighted FE"))+
  scale_linetype_discrete(name  ="Methods",
                          breaks=c("FE-Het", "hybrid_m", "hybrid_p", "FE-IPTW(pr)"),
                          labels=c("Fixed effects", 
                                   expression('Hybrid'['m']), expression('Hybrid'['p']), "Weighted FE"))+
  scale_shape_discrete(name  ="Methods",
                       breaks=c("FE-Het", "hybrid_m", "hybrid_p", "FE-IPTW(pr)"),
                       labels=c("Fixed effects", 
                                expression('Hybrid'['m']), expression('Hybrid'['p']), "Weighted FE"))



##Now compare variances of methods
var.val.mat<-v2_out_save%>%group_by(alpha,beta, rho, method) %>%
  summarise(var_var = var(var_est) )
var.est.frame<-merge(var.val.mat, var.est.frame, by=c("alpha", "beta", "rho", "method"))
var.est.frame$SD_of_var_est<-sqrt(var.est.frame$var_var)
var.est.frame.var<-var.est.frame[which(var.est.frame$method=="hybrid_m"|var.est.frame$method=="hybrid_p"|var.est.frame$method=="FE-Het"|var.est.frame$method=="FE-IPTW(pr)"),]

theme_set( theme_tufte() )
ggplot(var.est.frame.var, aes( x=beta, y=SD_of_var_est, col= method ) ) +
  facet_grid( alpha ~ rho, labeller = label_both) +
  geom_hline( yintercept=0, col="grey" ) +
  geom_line(size=0.4) + geom_point(size=0.8)+
  labs( y = "SD of of Variance Esitmators", x ="SD of Block Average Treatment Effects" )+
  scale_color_hue(name  ="Methods",
                  breaks=c("FE-Het", "hybrid_m", "hybrid_p", "FE-IPTW(pr)"),
                  labels=c("Fixed effects", expression('Hybrid'['m']), 
                           expression('Hybrid'['p']), "Weighted FE")) +
  theme(legend.position="bottom")


# Aggregate results to collapse the facets in the above.
s.var.est = var.est.frame.var %>%
  group_by( method, beta ) %>%
  dplyr::summarize( n = n(),
             avg.SD = mean(SD_of_var_est) )
ggplot(s.var.est, aes( x=beta, y=avg.SD, col= method ) ) +
  geom_hline( yintercept=0, col="grey" ) +
  geom_line(size=0.4) + geom_point(size=0.8)+
  labs( y = "SD of of Variance Esitmators", x ="SD of Block Average Treatment Effects" )+
  scale_color_hue(name  ="Methods",
                  breaks=c("FE-Het", "hybrid_m", "hybrid_p", "FE-IPTW(pr)"),
                  labels=c("Fixed effects", expression('Hybrid'['m']), 
                           expression('Hybrid'['p']), "Weighted FE")) +
  theme(legend.position="bottom")
